/*
===============================================================================

  FILE:  stega.cpp
  
  CONTENTS:
  
    A simple algorithm that encodes/decodes a random bitstring as a permutation
		of elements from a given reference ordering in O(n) time and O(n) space.
    This code is kept as simple as possible. It should be trivial to adapt the
    encode and decode routines to work with your own data elements.
    The "reference ordering" used here is obvioulsy not what you would usually
    use. Any order that the decoder can uniquely reconstruct will work.
  
  PROGRAMMERS:
  
    martin isenburg@cs.unc.edu
  
  COPYRIGHT:
  
    copyright (C) 2007 martin isenburg@cs.unc.edu
    
    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  
  CHANGE HISTORY:
  
    02 January 2008 -- fixed up for release
    15 October 2007 -- created after noticing that we do not need the binary
                       search - and the resulting additional O(logn) factor -
                       for a simple permutation steganography algorithm
  
===============================================================================
*/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <time.h>

class Permute
{
public:
	int p;
};

class Vertex : public Permute
{
public:
// not needed in this code
//	float coords[3];
};

class Face : public Permute
{
public:
// not needed in this code
//  int vertices[3];
};

static int bits_in_bit_buffer1;
static unsigned int bit_buffer1;
static unsigned int bit_buffer2;
static unsigned int bit_count;

static void init_bits()
{
	srand(0);
	bits_in_bit_buffer1 = 32;
	bit_buffer1 = ((rand()&2047) << 21) | ((rand()&2047) << 10) | (rand()&1023);
	bit_buffer2 = ((rand()&2047) << 21) | ((rand()&2047) << 10) | (rand()&1023);
	bit_count = 0;
}

static unsigned int peek_bits(int bits)
{
	int result;
	assert(bits <= 32 && bits_in_bit_buffer1);
	if (bits_in_bit_buffer1 >= bits)
	{
		result = bit_buffer1 & ((1<<bits)-1);
	}
	else
	{
		result = bit_buffer1 | ((bit_buffer2 & ((1<<(bits-bits_in_bit_buffer1))-1)) << bits_in_bit_buffer1);
	}
	return result;
}

static void pop_bits(int bits)
{
	assert(bits <= 32 && bits_in_bit_buffer1);
	if (bits_in_bit_buffer1 > bits)
	{
		bits_in_bit_buffer1 -= bits;
		bit_buffer1 >>= bits;
	}
	else
	{
		bits_in_bit_buffer1 = bits_in_bit_buffer1-bits+32;
		bit_buffer1 = bit_buffer2 >> (32-bits_in_bit_buffer1);
		bit_buffer2 = ((rand()&2047) << 21) | ((rand()&2047) << 10) | (rand()&1023);
	}
	bit_count += bits;
}

static int encode(Permute** reference, Permute** permutation, int nelements)
{
	int bit_count_start = bit_count;

	// compute k and when k will decrement
	int i = nelements;
	int k = 0;
	while (i/=2) k++;
	int twopowk = 1<<k;
	
	// encode either k+1 or k bits with each element
	int p;
  for (i = 0; i < nelements-1; i++)
	{
		p = peek_bits(k+1);
		if (p >= twopowk && p < nelements-i)
		{
			// we were able to encode k+1 bits
			pop_bits(k+1);
		}
		else
		{
			// can only encode k bits
			p = peek_bits(k);
			pop_bits(k);
		}
		// we encode them by putting the reference element p in the output
		permutation[i] = reference[p];
    // this line replaces the binary tree
		// the last element of the reference array becomes the new reference element p 
		reference[p] = reference[nelements-1-i];
		// is k becomming smaller
		if (nelements-i == twopowk)
		{
			k--;
			twopowk>>=1;
		}
	}

	// we can not encode anything with the very last element 
	permutation[nelements-1] = reference[0];

	// return number of bits encoded
	return bit_count - bit_count_start;
}

static int decode(Permute** reference, Permute** permutation, int nelements)
{
	int bit_count_start = bit_count;

	// compute k and when k will decrement
	int i = nelements;
	int k = 0;
	while (i/=2) k++;
	int twopowk = 1<<k;

	// decode either k+1 or k bits with each face
	int p, p_orig;
  for (i = 0; i < nelements-1; i++)
	{
		p = permutation[i]->p;
		if (p >= twopowk)
		{
			// we were able to decode k+1 bits
			p_orig = peek_bits(k+1);
			pop_bits(k+1);
		}
		else
		{
			// can only decode k bits
			p_orig = peek_bits(k);
			pop_bits(k);
		}
		// check that we decoded the right number
		if (p != p_orig)
		{
			fprintf(stderr,"decoded %d instead of %d\n",p,p_orig);
		}
    // these two lines replace the binary tree
		// the last face of the reference array becomes the new reference face p 
		reference[p] = reference[nelements-1-i];
		reference[p]->p = p;
		// is k becomming smaller
		if (nelements-i == twopowk)
		{
			k--;
			twopowk>>=1;
		}
	}
	// we can not decode anything from the very last face 
	if (permutation[nelements-1]->p != 0)
	{
		fprintf(stderr,"last face has position %d instead of 0\n",permutation[nelements-1]->p);
	}

	// return number of bits decoded
	return bit_count - bit_count_start;
}

int main(int argc, char **argv)
{
  int i;
//  int nverts = 2832, nfaces = 5660;
//  int nverts = 34834, nfaces = 69451;
//  int nverts = 138490, nfaces = 277384;
  int nverts = 437500, nfaces = 872252;

  clock_t clock_start, clock_finish;

  if (argc == 3)
  {
		nverts = atoi(argv[1]);
		nfaces = atoi(argv[2]);
  }
  else if (argc != 1)
  {
    fprintf(stderr,"usage:\n");
    fprintf(stderr,"%s <nverts> <nfaces>\n", argv[0]);
    exit(1);
  }

  // create (aka "load") the mesh
  Vertex* vertices = new Vertex[nverts];
  Face* faces = new Face[nfaces];

	// create "some" reference ordering. in the actual coder that would
	// involve reconstructing the mesh connectivity and then traversing it 
	// with some deterministic traversal starting from an agreed-upon 
	// oriented seed edge such as a topological breadth-first or depth-first 
	// traversal such as Keeler-Westbrook 1996 or Rossignac 1999 or the (just
	// slightly different) traversal of the TG coder 1998.
  Permute** verts_reference = new Permute*[nverts];
  for (i = 0; i < nverts; i++)
  {
	  verts_reference[i] = &(vertices[(i+253)%nverts]);
  }
  Permute** faces_reference = new Permute*[nfaces];
  for (i = 0; i < nfaces; i++)
  {
	  faces_reference[i] = &(faces[(i+57)%nfaces]);
  }

	// create arrays to store the computed permutation
  Permute** verts_permutation = new Permute*[nverts];
  Permute** faces_permutation = new Permute*[nfaces];

	// prepare the bits that are encoded
	init_bits();

  // start clock
  clock_start = clock();

	// encode as many bits as possible by permuting the vertices
	i = encode(verts_reference, verts_permutation, nverts);
	fprintf(stderr,"encoded %d bits in %d vertices (avg %.2f bits)\n",i,nverts,((float)i)/((float)nverts));

	// encode as many bits as possible by permuting the faces
	i = encode(faces_reference, faces_permutation, nfaces);
	fprintf(stderr,"encoded %d bits in %d faces (avg %.2f bits)\n",i,nfaces,((float)i)/((float)nfaces));

  // stop clock
  clock_finish = clock();

	// done encoding
	fprintf(stderr,"done. message of %d bits hidden in connectivity (%.2f b/v)\n",bit_count,((float)bit_count)/((float)nverts));
  fprintf(stderr,"      hiding message in mesh took %.3f seconds\n", (double)(clock_finish-clock_start)/CLOCKS_PER_SEC);

	// start decoding

	// re-create the same reference orderings that were created above
  for (i = 0; i < nverts; i++)
  {
	  verts_reference[i] = &(vertices[(i+253)%nverts]);
		verts_reference[i]->p = i;
  }
  for (i = 0; i < nfaces; i++)
  {
	  faces_reference[i] = &(faces[(i+57)%nfaces]);
		faces_reference[i]->p = i;
  }

	// re-prepare the bits to check that we decode correctly
	init_bits();

  // start clock
  clock_start = clock();

	// decode bits from vertex permutation
	i = decode(verts_reference, verts_permutation, nverts);
	fprintf(stderr,"decoded %d bits from %d vertices\n",i,nverts);

	// decode bits from face permutation
	i = decode(faces_reference, faces_permutation, nfaces);
	fprintf(stderr,"decoded %d bits from %d faces\n",i,nfaces);

  // stop clock
  clock_finish = clock();

	// done decoding
	fprintf(stderr,"done. message of %d bits recovered from connectivity\n",bit_count);
  fprintf(stderr,"      recovering and checking message took %.3f seconds\n", (double)(clock_finish-clock_start)/CLOCKS_PER_SEC);

  return 1;
}
